/*
 * kgraft_patch_bsc962078
 *
 * Fix for CVE-2016-0728, bsc#962078
 *
 *  Upstream commit:
 *  23567fd052a9abb6d67fe8e7a9ccdd9800a540f2
 *
 *  SLE12 commit:
 *  918e5764e2e8b05f486241b3eb2d4f95b5b142c3
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/sched.h>
#include <linux/cred.h>

static int (*kgr_install_session_keyring_to_cred)(struct cred *cred, struct key *keyring);
static struct mutex *kgr_key_session_mutex;
static struct key *(*kgr_find_keyring_by_name)(const char *name, bool skip_perm_check);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "install_session_keyring_to_cred", (void *)&kgr_install_session_keyring_to_cred },
	{ "key_session_mutex", (void *)&kgr_key_session_mutex },
	{ "find_keyring_by_name", (void *)&kgr_find_keyring_by_name },
};

/* Patched function */
long kgr_join_session_keyring(const char *name)
{
	const struct cred *old;
	struct cred *new;
	struct key *keyring;
	long ret, serial;

	new = prepare_creds();
	if (!new)
		return -ENOMEM;
	old = current_cred();

	/* if no name is provided, install an anonymous keyring */
	if (!name) {
		ret = kgr_install_session_keyring_to_cred(new, NULL);
		if (ret < 0)
			goto error;

		serial = new->session_keyring->serial;
		ret = commit_creds(new);
		if (ret == 0)
			ret = serial;
		goto okay;
	}

	/* allow the user to join or create a named keyring */
	mutex_lock(kgr_key_session_mutex);

	/* look for an existing keyring of this name */
	keyring = kgr_find_keyring_by_name(name, false);
	if (PTR_ERR(keyring) == -ENOKEY) {
		/* not found - try and create a new one */
		keyring = keyring_alloc(
			name, old->uid, old->gid, old,
			KEY_POS_ALL | KEY_USR_VIEW | KEY_USR_READ | KEY_USR_LINK,
			KEY_ALLOC_IN_QUOTA, NULL);
		if (IS_ERR(keyring)) {
			ret = PTR_ERR(keyring);
			goto error2;
		}
	} else if (IS_ERR(keyring)) {
		ret = PTR_ERR(keyring);
		goto error2;
	} else if (keyring == new->session_keyring) {
		/*
		 * Fix CVE-2016-0728
		 *  +next row
		 */
		key_put(keyring);
		ret = 0;
		goto error2;
	}

	/* we've got a keyring - now to install it */
	ret = kgr_install_session_keyring_to_cred(new, keyring);
	if (ret < 0)
		goto error2;

	commit_creds(new);
	mutex_unlock(kgr_key_session_mutex);

	ret = keyring->serial;
	key_put(keyring);
okay:
	return ret;

error2:
	mutex_unlock(kgr_key_session_mutex);
error:
	abort_creds(new);
	return ret;
}

static int kgr_patch_bsc962078_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_bsc962078_init(void)
{
	return kgr_patch_bsc962078_kallsyms();
}
